@@ -19,6 +19,8 @@
 
 import collections
 import difflib
+import functools
+import gc
 import ipaddress
 import json
 import hashlib
@@ -79,6 +81,8 @@ def read() -> dict:
 def write(snitch: dict) -> None:
     """write snitch to correct location (root privileges should be dropped first)"""
     file_path = os.path.join(os.path.expanduser("~"), ".config", "picosnitch", "snitch.json")
+    if snitch.pop("WRITELOCK", False):
+        file_path += "~"
     if not os.path.isdir(os.path.dirname(file_path)):
         os.makedirs(os.path.dirname(file_path))
     try:
@@ -95,10 +99,21 @@ def drop_root_privileges() -> None:
         os.setuid(int(os.getenv("SUDO_UID")))
 
 
-def terminate(snitch: dict, p_snitch_mon: multiprocessing.Process, q_term: multiprocessing.Queue, q_vt_term: multiprocessing.Queue):
+def terminate(snitch: dict, p_snitch_mon: multiprocessing.Process,
+              q_error: multiprocessing.Queue,
+              q_term: multiprocessing.Queue,
+              q_sha_term: multiprocessing.Queue,
+              q_psutil_term: multiprocessing.Queue,
+              q_vt_term: multiprocessing.Queue):
     """write snitch one last time, then terminate picosnitch and subprocesses if running"""
+    while not q_error.empty():
+        error = q_error.get()
+        snitch["Errors"].append(time.ctime() + " " + error)
+        toast(error, file=sys.stderr)
     write(snitch)
     q_vt_term.put("TERMINATE")
+    q_psutil_term.put("TERMINATE")
+    q_sha_term.put("TERMINATE")
     q_term.put("TERMINATE")
     p_snitch_mon.join(5)
     p_snitch_mon.close()
@@ -149,12 +164,21 @@ def get_sha256(exe: str) -> str:
     """get sha256 of process executable"""
     try:
         with open(exe, "rb") as f:
-            sha256 = hashlib.sha256(f.read())
-        return sha256.hexdigest()
+            sha256 = hashlib.sha256(f.read()).hexdigest()
+        return sha256
     except Exception:
         return "0000000000000000000000000000000000000000000000000000000000000000"
 
 
+def get_proc_info(pid: int) -> typing.Union[dict, None]:
+    """use psutil to get proc info from pid"""
+    try:
+        proc = psutil.Process(pid).as_dict(attrs=["name", "exe", "cmdline", "pid"], ad_value="")
+    except Exception:
+        proc = None
+    return proc
+
+
 def get_vt_results(snitch: dict, q_vt: multiprocessing.Queue, check_pending: bool = False):
     """get virustotal results from subprocess and update snitch"""
     if check_pending:
@@ -171,24 +195,27 @@ def get_vt_results(snitch: dict, q_vt: multiprocessing.Queue, check_pending: boo
                 toast("Suspicious results for " + proc["name"])
 
 
-def initial_poll(snitch: dict, known_pids: dict, q_vt_pending: multiprocessing.Queue) -> None:
-    """poll initial processes and connections using psutil, then runs update_snitch_*"""
+def initial_poll(snitch: dict, known_pids: dict) -> list:
+    """poll initial processes and connections using psutil and queue for update_snitch()"""
     ctime = time.ctime()
+    update_snitch_pending = []
     current_processes = {}
     for proc in psutil.process_iter(attrs=["name", "exe", "cmdline", "pid"], ad_value=""):
-        if os.path.isfile(proc.info["exe"]):
-            current_processes[proc.info["exe"]] = proc.info
-            known_pids["pid"] = proc.info
+        proc = proc.info
+        if os.path.isfile(proc["exe"]):
+            proc["cmdline"] = shlex.join(proc["cmdline"])
+            current_processes[proc["exe"]] = proc
+            known_pids[proc["pid"]] = proc
     proc = {"name": "", "exe": "", "cmdline": "", "pid": ""}
     current_connections = set(psutil.net_connections(kind="all"))
     for conn in current_connections:
         try:
             if conn.pid is not None and conn.raddr and not ipaddress.ip_address(conn.raddr.ip).is_private:
                 proc = psutil.Process(conn.pid).as_dict(attrs=["name", "exe", "cmdline", "pid"], ad_value="")
+                proc["cmdline"] = shlex.join(proc["cmdline"])
                 conn_dict = {"ip": conn.raddr.ip, "port": conn.raddr.port}
-                sha256 = get_sha256(proc["exe"])
                 _ = current_processes.pop(proc["exe"], 0)
-                update_snitch(snitch, proc, conn_dict, sha256, ctime, q_vt_pending)
+                update_snitch_pending.append((proc, conn_dict, ctime))
         except Exception as e:
             # too late to grab process info (most likely) or some other error
             error = "Init " + type(e).__name__ + str(e.args) + str(conn)
@@ -198,27 +225,31 @@ def initial_poll(snitch: dict, known_pids: dict, q_vt_pending: multiprocessing.Q
                 error += "{process no longer exists}"
             snitch["Errors"].append(ctime + " " + error)
     if not snitch["Config"]["Only log connections"]:
-        conn = {"ip": "", "port": 0}
+        conn = {"ip": "", "port": -1}
         for proc in current_processes.values():
-            try:
-                sha256 = get_sha256(proc["exe"])
-                update_snitch(snitch, proc, conn, sha256, ctime, q_vt_pending)
-            except Exception as e:
-                error = "Init " + type(e).__name__ + str(e.args) + str(proc)
-                snitch["Errors"].append(ctime + " " + error)
+            update_snitch_pending.append((proc, conn, ctime))
+    return update_snitch_pending
 
 
-def process_queue(snitch: dict, known_pids: dict, missed_conns: list, new_processes: list, q_vt_pending: multiprocessing.Queue) -> list:
-    """process list of new processes and call update_snitch"""
+def process_queue(snitch: dict, known_pids: dict, missed_conns: list, new_processes: list,
+                  q_psutil_pending: multiprocessing.Queue,
+                  q_psutil_results: multiprocessing.Queue
+                  ) -> tuple:
+    """process list of new processes and queue for update_snitch()"""
     ctime = time.ctime()
+    update_snitch_pending = []
     pending_list = []
     pending_conns = []
     for proc in new_processes:
         try:
             if proc["type"] == "exec":
-                proc["exe"] = shlex.split(proc["cmdline"])[0]
+                try:
+                    cmdline = shlex.split(proc["cmdline"])
+                except Exception:
+                    cmdline = proc["cmdline"].strip().split()
+                proc["exe"] = cmdline[0]
                 if proc["exe"] == "exec":
-                    proc["exe"] = shlex.split(proc["cmdline"])[1]
+                    proc["exe"] = cmdline[1]
                 known_pids[proc["pid"]] = proc
                 if not snitch["Config"]["Only log connections"]:
                     pending_list.append(proc)
@@ -229,9 +260,27 @@ def process_queue(snitch: dict, known_pids: dict, missed_conns: list, new_proces
                     proc["cmdline"] = known_pids[proc["pid"]]["cmdline"]
                     pending_list.append(proc)
                 else:
-                    proc_psutil = psutil.Process(proc["pid"]).as_dict(attrs=["name", "exe", "cmdline", "pid"], ad_value="")
-                    if proc_psutil["exe"]:
-                        known_pids[proc_psutil["pid"]] = proc_psutil
+                    try:
+                        q_psutil_pending.put(pickle.dumps(proc["pid"]))
+                        proc_psutil = pickle.loads(q_psutil_results.get())
+                        if proc_psutil["exe"]:
+                            proc_psutil["cmdline"] = shlex.join(proc_psutil["cmdline"])
+                            known_pids[proc_psutil["pid"]] = proc_psutil
+                    except Exception:
+                        if proc["ppid"] not in known_pids:
+                            try:
+                                q_psutil_pending.put(pickle.dumps(proc["ppid"]))
+                                proc_psutil = pickle.loads(q_psutil_results.get())
+                                if proc_psutil["exe"]:
+                                    proc_psutil["cmdline"] = shlex.join(proc_psutil["cmdline"])
+                                    known_pids[proc_psutil["pid"]] = proc_psutil
+                                    # use this as best guess for child too since it probably forked and was short lived
+                                    # if not, exec should have caught it on the next round
+                                    # this is why pending conns is now postponed till the next iteration
+                                    known_pids[proc["pid"]] = proc_psutil
+                            except Exception:
+                                pass
+                    proc["missed"] = 1
                     pending_conns.append(proc)
         except Exception as e:
             error = "Process queue " + type(e).__name__ + str(e.args) + str(proc)
@@ -242,25 +291,44 @@ def process_queue(snitch: dict, known_pids: dict, missed_conns: list, new_proces
             proc["exe"] = known_pids[proc["pid"]]["exe"]
             proc["cmdline"] = known_pids[proc["pid"]]["cmdline"]
             pending_list.append(proc)
+        elif proc["missed"] < 5:
+            # give it 5 rounds to find the pid in exec (2 should be enough)
+            # don't waste cpu checking psutil anymore since it would be short lived
+            # therefore ppid is also unlikely to appear (less harm in checking this maybe once more)
+            proc["missed"] += 1
+            pending_conns.append(proc)
         else:
+            _ = proc.pop("missed")
             snitch["Errors"].append(ctime + " no known process for conn: " + str(proc))
     for proc in pending_list:
+        if proc["type"] == "conn":
+            conn = {"ip": proc["ip"], "port": proc["port"]}
+        else:
+            conn = {"ip": "", "port": -1}
+        update_snitch_pending.append((proc, conn, ctime))
+    return pending_conns, update_snitch_pending
+
+
+def update_snitch_wrapper(snitch: dict, update_snitch_pending: list,
+                          q_sha_pending: multiprocessing.Queue,
+                          q_sha_results: multiprocessing.Queue,
+                          q_vt_pending: multiprocessing.Queue):
+    """loop over update_snitch() with pending update list"""
+    for proc, conn, ctime in update_snitch_pending:
         try:
-            if proc["type"] == "conn":
-                conn = {"ip": proc["ip"], "port": proc["port"]}
-            else:
-                conn = {"ip": "", "port": 0}
-            sha256 = get_sha256(proc["exe"])
+            q_sha_pending.put(pickle.dumps(proc["exe"]))
+            sha256 = pickle.loads(q_sha_results.get())
             update_snitch(snitch, proc, conn, sha256, ctime, q_vt_pending)
         except Exception as e:
             error = "Update snitch " + type(e).__name__ + str(e.args) + str(proc)
             snitch["Errors"].append(ctime + " " + error)
             toast("Update snitch error: " + error, file=sys.stderr)
-    return pending_conns
 
 
 def update_snitch(snitch: dict, proc: dict, conn: dict, sha256: str, ctime: str, q_vt_pending: multiprocessing.Queue) -> None:
     """update the snitch with data from queues and create a notification if new entry"""
+    # Prevent overwriting the snitch before this function completes in the event of a termination signal
+    snitch["WRITELOCK"] = True
     # Get DNS reverse name and reverse the name for sorting
     reversed_dns = reverse_domain_name(reverse_dns_lookup(conn["ip"]))
     # Omit fields from log
@@ -276,14 +344,16 @@ def update_snitch(snitch: dict, proc: dict, conn: dict, sha256: str, ctime: str,
         if proc["exe"] not in snitch["Names"][proc["name"]]:
             snitch["Names"][proc["name"]].append(proc["exe"])
             toast("New executable detected for " + proc["name"] + ": " + proc["exe"])
-    elif conn["ip"] or conn["port"]:
+    elif conn["ip"] or conn["port"] >= 0:  # port 0 is a conn where port wasn't detected, -1 is proc without conn detected
         snitch["Names"][proc["name"]] = [proc["exe"]]
         toast("First network connection detected for " + proc["name"])
+    elif not snitch["Config"]["Only log connections"]:
+        snitch["Names"][proc["name"]] = [proc["exe"]]
     # Update Processes
     if proc["exe"] not in snitch["Processes"]:
         snitch["Processes"][proc["exe"]] = {
             "name": proc["name"],
-            "cmdlines": [str(proc["cmdline"])],
+            "cmdlines": [proc["cmdline"]],
             "first seen": ctime,
             "last seen": ctime,
             "days seen": 1,
@@ -298,8 +368,8 @@ def update_snitch(snitch: dict, proc: dict, conn: dict, sha256: str, ctime: str,
         entry = snitch["Processes"][proc["exe"]]
         if proc["name"] not in entry["name"]:
             entry["name"] += " alternative=" + proc["name"]
-        if str(proc["cmdline"]) not in entry["cmdlines"]:
-            get_common_pattern(str(proc["cmdline"]), entry["cmdlines"], 0.8)
+        if proc["cmdline"] not in entry["cmdlines"]:
+            get_common_pattern(proc["cmdline"], entry["cmdlines"], 0.8)
             entry["cmdlines"].sort()
         if conn["port"] not in entry["ports"]:
             entry["ports"].append(conn["port"])
@@ -323,6 +393,8 @@ def update_snitch(snitch: dict, proc: dict, conn: dict, sha256: str, ctime: str,
     else:
         if conn["port"] not in snitch["Config"]["Remote address unlog"] and proc["name"] not in snitch["Config"]["Remote address unlog"]:
             snitch["Remote Addresses"][reversed_dns] = ["First connection: " + ctime, proc["exe"]]
+    # Unlock the snitch for writing
+    _ = snitch.pop("WRITELOCK")
 
 
 def loop(vt_api_key: str = ""):
@@ -335,14 +407,18 @@ def loop(vt_api_key: str = ""):
     _ = snitch.pop("Template", 0)
     if vt_api_key:
         snitch["Config"]["VT API key"] = vt_api_key
-    # init bpf program
-    p_snitch_mon, q_snitch, q_error, q_term = init_snitch_subprocess(snitch["Config"])
+    # init root subprocesses and do initial poll with root before dropping privileges
+    p_sha, q_sha_pending, q_sha_results, q_sha_term = init_func_subprocess("snitchsha", functools.lru_cache(get_sha256))
+    p_psutil, q_psutil_pending, q_psutil_results, q_psutil_term = init_func_subprocess("snitchpsutil", get_proc_info)
+    p_snitch_mon, q_snitch, q_error, q_term = init_snitch_subprocess(snitch["Config"], p_sha, p_psutil)
+    known_pids = collections.OrderedDict()
+    update_snitch_pending = initial_poll(snitch, known_pids)
     drop_root_privileges()
-    # init virustotal subprocess
+    # init subprocesses without root (just virustotal)
     p_virustotal, q_vt_pending, q_vt_results, q_vt_term = init_virustotal_subprocess(snitch["Config"])
     # set signal handlers
-    signal.signal(signal.SIGTERM, lambda *args: terminate(snitch, p_snitch_mon, q_term, q_vt_term))
-    signal.signal(signal.SIGINT, lambda *args: terminate(snitch, p_snitch_mon, q_term, q_vt_term))
+    signal.signal(signal.SIGTERM, lambda *args: terminate(snitch, p_snitch_mon, q_error, q_term, q_sha_term, q_psutil_term, q_vt_term))
+    signal.signal(signal.SIGINT, lambda *args: terminate(snitch, p_snitch_mon, q_error, q_term, q_sha_term, q_psutil_term, q_vt_term))
     # snitch init checks
     if p_snitch_mon is None:
         snitch["Errors"].append(time.ctime() + " Snitch subprocess init failed, __name__ != __main__, try: sudo -E python -m picosnitch")
@@ -351,13 +427,13 @@ def loop(vt_api_key: str = ""):
     # check if there are pending virtustotal results from last time
     get_vt_results(snitch, q_vt_pending, True)
     # init variables for loop
-    known_pids = {}
     missed_conns = []
     sizeof_snitch = sys.getsizeof(pickle.dumps(snitch))
     last_write = 0
-    # get initial running processes and connections
-    initial_poll(snitch, known_pids, q_vt_pending)
+    # update snitch with initial running processes and connections
+    update_snitch_wrapper(snitch, update_snitch_pending, q_sha_pending, q_sha_results, q_vt_pending)
     known_pids[p_virustotal.pid] = {"name": p_virustotal.name, "exe": __file__, "cmdline": shlex.join(sys.argv), "pid": p_virustotal.pid}
+    # the actual main loop
     while True:
         # check for subprocess errors
         while not q_error.empty():
@@ -368,8 +444,8 @@ def loop(vt_api_key: str = ""):
         if not p_snitch_mon.is_alive():
             snitch["Errors"].append(time.ctime() + " snitch subprocess stopped")
             toast("snitch subprocess stopped, exiting picosnitch", file=sys.stderr)
-            terminate(snitch, p_snitch_mon, q_term, q_vt_term)
-        # list of new processes and connections since last poll
+            terminate(snitch, p_snitch_mon, q_error, q_term, q_sha_term, q_psutil_term, q_vt_term)
+        # get list of new processes and connections since last update
         time.sleep(5)
         new_processes = []
         try:
@@ -377,10 +453,17 @@ def loop(vt_api_key: str = ""):
                 new_processes.append(q_snitch.get(block=False))
         except queue.Empty:
             pass
+        # process the list and update snitch
         new_processes = [pickle.loads(proc) for proc in new_processes]
-        missed_conns = process_queue(snitch, known_pids, missed_conns, new_processes, q_vt_pending)
-        # update snitch with virtustotal results
+        missed_conns, update_snitch_pending = process_queue(snitch, known_pids, missed_conns, new_processes, q_psutil_pending, q_psutil_results)
+        update_snitch_wrapper(snitch, update_snitch_pending, q_sha_pending, q_sha_results, q_vt_pending)
         get_vt_results(snitch, q_vt_results, False)
+        # free some memory
+        while len(known_pids) > 9000:
+            _ = known_pids.popitem(last=False)
+        del(new_processes)
+        del(update_snitch_pending)
+        gc.collect()
         # write snitch
         if time.time() - last_write > 30:
             new_size = sys.getsizeof(pickle.dumps(snitch))
@@ -390,24 +473,30 @@ def loop(vt_api_key: str = ""):
                 write(snitch)
 
 
-def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process, multiprocessing.Queue, multiprocessing.Queue, multiprocessing.Queue]:
-    """init snitch subprocess and monitor with root (before dropping root privileges)"""
+def init_snitch_subprocess(config: dict, p_sha, p_psutil) -> typing.Tuple[multiprocessing.Process, multiprocessing.Queue, multiprocessing.Queue, multiprocessing.Queue]:
+    """init snitch subprocesses and monitor with root (before dropping root privileges)"""
     def snitch_linux(config, q_snitch, q_error, q_term_monitor):
         """runs a bpf program to monitor the system for new processes and connections and puts them in the queue"""
         from bcc import BPF
         if os.getuid() == 0:
+            last_data = {"pid": 0, "ppid": 0}
             b = BPF(text=bpf_text)
             execve_fnname = b.get_syscall_fnname("execve")
             b.attach_kprobe(event=execve_fnname, fn_name="syscall__execve")
             b.attach_kretprobe(event=execve_fnname, fn_name="do_ret_sys_execve")
             b.attach_kprobe(event="security_socket_connect", fn_name="security_socket_connect_entry")
+            b.attach_uprobe(name="c", sym="getaddrinfo", fn_name="do_entry", pid=-1)
+            b.attach_uprobe(name="c", sym="gethostbyname", fn_name="do_entry", pid=-1)
+            b.attach_uprobe(name="c", sym="gethostbyname2", fn_name="do_entry", pid=-1)
+            b.attach_uretprobe(name="c", sym="getaddrinfo", fn_name="do_return", pid=-1)
+            b.attach_uretprobe(name="c", sym="gethostbyname", fn_name="do_return", pid=-1)
+            b.attach_uretprobe(name="c", sym="gethostbyname2", fn_name="do_return", pid=-1)
             argv = collections.defaultdict(list)
             def queue_exec_event(cpu, data, size):
                 event = b["exec_events"].event(data)
                 if event.type == 0:  # EVENT_ARG
                     argv[event.pid].append(event.argv)
                 elif event.type == 1:  # EVENT_RET
-                    argv[event.pid] = [b"\"" + arg.replace(b"\"", b"\\\"") + b"\"" for arg in argv[event.pid]]
                     argv_text = b' '.join(argv[event.pid]).replace(b'\n', b'\\n')
                     q_snitch.put(pickle.dumps({"type": "exec", "pid": event.pid, "name": event.comm.decode(), "cmdline": argv_text.decode()}))
                     try:
@@ -416,17 +505,21 @@ def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process
                         pass
             def queue_ipv4_event(cpu, data, size):
                 event = b["ipv4_events"].event(data)
-                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "port": event.dport, "ip": socket.inet_ntop(socket.AF_INET, struct.pack("I", event.daddr))}))
+                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "ppid": event.ppid, "name": event.task.decode(), "port": event.dport, "ip": socket.inet_ntop(socket.AF_INET, struct.pack("I", event.daddr))}))
             def queue_ipv6_event(cpu, data, size):
                 event = b["ipv6_events"].event(data)
-                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "port": event.dport, "ip": socket.inet_ntop(socket.AF_INET6, event.daddr)}))
+                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "ppid": event.ppid, "name": event.task.decode(), "port": event.dport, "ip": socket.inet_ntop(socket.AF_INET6, event.daddr)}))
             def queue_other_event(cpu, data, size):
                 event = b["other_socket_events"].event(data)
-                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "port": 0, "ip": ""}))
+                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "ppid": event.ppid, "name": event.task.decode(), "port": 0, "ip": ""}))
+            def queue_dns_event(cpu, data, size):
+                event = b["dns_events"].event(data)
+                q_snitch.put(pickle.dumps({"type": "conn", "pid": event.pid, "ppid": event.ppid, "name": event.comm.decode(), "port": 0, "ip": "", "host": event.host.decode()}))
             b["exec_events"].open_perf_buffer(queue_exec_event)
             b["ipv4_events"].open_perf_buffer(queue_ipv4_event)
             b["ipv6_events"].open_perf_buffer(queue_ipv6_event)
             b["other_socket_events"].open_perf_buffer(queue_other_event)
+            b["dns_events"].open_perf_buffer(queue_dns_event)
             while True:
                 try:
                     b.perf_buffer_poll()
@@ -443,7 +536,7 @@ def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process
             q_error.put("Snitch subprocess permission error, requires root")
         return 1
 
-    def snitch_monitor(config, q_snitch, q_error, q_term):
+    def snitch_monitor(config, q_snitch, q_error, q_term, p_sha, p_psutil):
         """monitor the snitch subprocess and parent process, has same privileges as subprocess for clean termination at the command or death of the parent"""
         signal.signal(signal.SIGINT, lambda *args: None)
         if sys.platform.startswith("linux"):
@@ -453,6 +546,7 @@ def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process
         else:
             q_error.put("Did not detect a supported operating system")
             return 1
+        p_sha, p_psutil = psutil.Process(p_sha.pid), psutil.Process(p_psutil.pid)
         q_term_monitor = multiprocessing.Queue()
         terminate_subprocess = lambda p_snitch_sub: q_term_monitor.put("TERMINATE") or p_snitch_sub.join(3) or (p_snitch_sub.is_alive() and p_snitch_sub.kill()) or p_snitch_sub.close()
         p_snitch_sub = multiprocessing.Process(name="snitchsubprocess", target=p_snitch_func, args=(config, q_snitch, q_error, q_term_monitor), daemon=True)
@@ -471,6 +565,16 @@ def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process
                 p_snitch_sub = multiprocessing.Process(name="snitchsubprocess", target=p_snitch_func, args=(config, q_snitch, q_error, q_term_monitor), daemon=True)
                 p_snitch_sub.start()
                 time_last_start = time.time()
+            if not (p_sha.is_running() and p_sha.status() != "zombie" and p_psutil.is_running() and p_psutil.status() != "zombie"):
+                # if either of these die, the main process will hang on the next request to either of them
+                # better to just take everything down so the user can see it stopped running and doesn't get the impression that everything is still working
+                # the only way this temporary solution should hang is if:
+                # a) the main process hangs and this monitor is killed before the 10 second timeout
+                # b) this monitor is killed then p_sha or p_psutil is killed causing the main process to hang before it checks if the monitor is still alive
+                # todo: handle this better at some point (basically restructure everything)
+                q_error.put("sha256 or psutil subprocess died, terminating picosnitch")
+                psutil.Process(multiprocessing.parent_process().pid).terminate()
+                break
             try:
                 if q_term.get(block=True, timeout=10):
                     break
@@ -482,12 +586,39 @@ def init_snitch_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process
 
     if __name__ == "__main__":
         q_snitch, q_error, q_term = multiprocessing.Queue(), multiprocessing.Queue(), multiprocessing.Queue()
-        p_snitch_mon = multiprocessing.Process(name="snitchsubprocessmonitor", target=snitch_monitor, args=(config, q_snitch, q_error, q_term))
+        p_snitch_mon = multiprocessing.Process(name="snitchmonitor", target=snitch_monitor, args=(config, q_snitch, q_error, q_term, p_sha, p_psutil))
         p_snitch_mon.start()
         return p_snitch_mon, q_snitch, q_error, q_term
     return None, None, None, None
 
 
+def init_func_subprocess(name: str, func: typing.Callable) -> typing.Tuple[multiprocessing.Process, multiprocessing.Queue, multiprocessing.Queue, multiprocessing.Queue]:
+    """init subprocess for function"""
+    def func_subprocess(func: typing.Callable, q_pending, q_results, q_term):
+        """wrapper function for subprocess"""
+        last_error = 0
+        while True:
+            try:
+                if not q_term.empty() or not multiprocessing.parent_process().is_alive():
+                    return 0
+                arg = pickle.loads(q_pending.get(block=True, timeout=15))
+                q_results.put(pickle.dumps(func(arg)))
+            except queue.Empty:
+                # have to timeout here to check whether to terminate otherwise this could stay hanging
+                # daemon=True flag for multiprocessing.Process does not work after root privileges are dropped for parent
+                pass
+            except Exception:
+                if time.time() - last_error < 30:
+                    return 1
+                last_error = time.time()
+    if __name__ == "__main__":
+        q_pending, q_results, q_term = multiprocessing.Queue(), multiprocessing.Queue(), multiprocessing.Queue()
+        p_func = multiprocessing.Process(name=name, target=func_subprocess, args=(func, q_pending, q_results, q_term))
+        p_func.start()
+        return p_func, q_pending, q_results, q_term
+    return None, None, None, None
+
+
 def init_virustotal_subprocess(config: dict) -> typing.Tuple[multiprocessing.Process, multiprocessing.Queue, multiprocessing.Queue, multiprocessing.Queue]:
     """init virustotal subprocess"""
     def virustotal_subprocess(config: dict, q_vt_pending, q_vt_results, q_vt_term):
@@ -525,11 +656,9 @@ def init_virustotal_subprocess(config: dict) -> typing.Tuple[multiprocessing.Pro
                 if time.time() - last_error < 30:
                     return 1
                 last_error = time.time()
-
-
     if __name__ == "__main__":
         q_vt_pending, q_vt_results, q_vt_term = multiprocessing.Queue(), multiprocessing.Queue(), multiprocessing.Queue()
-        p_virustotal = multiprocessing.Process(name="virustotalsnitch", target=virustotal_subprocess, args=(config, q_vt_pending, q_vt_results, q_vt_term), daemon=True)
+        p_virustotal = multiprocessing.Process(name="snitchvirustotal", target=virustotal_subprocess, args=(config, q_vt_pending, q_vt_results, q_vt_term), daemon=True)
         p_virustotal.start()
         return p_virustotal, q_vt_pending, q_vt_results, q_vt_term
     return None, None, None, None
@@ -567,7 +696,9 @@ def main():
 bpf_text = """
 // This BPF program was adapted from the following sources, both licensed under the Apache License, Version 2.0
 // https://github.com/iovisor/bcc/blob/023154c7708087ddf6c2031cef5d25c2445b70c4/tools/execsnoop.py
+// https://github.com/iovisor/bcc/blob/ab14fafec3fc13f89bd4678b7fc94829dcacaa7b/tools/gethostlatency.py
 // https://github.com/p-/socket-connect-bpf/blob/7f386e368759e53868a078570254348e73e73e22/securitySocketConnectSrc.bpf
+// https://github.com/p-/socket-connect-bpf/blob/7f386e368759e53868a078570254348e73e73e22/dnsLookupSrc.bpf
 // Copyright 2016 Netflix, Inc.
 // Copyright 2019 Peter Stöckli
 
@@ -591,6 +722,8 @@ bpf_text = """
 #include <linux/in6.h>
 #include <linux/ip.h>
 
+// execsnoop structs
+
 #define ARGSIZE  128
 
 enum event_type {
@@ -609,9 +742,12 @@ struct data_t {
 };
 BPF_PERF_OUTPUT(exec_events);
 
+// securitySocketConnect structs
+
 struct ipv4_event_t {
     u64 ts_us;
     u32 pid;
+    u32 ppid;
     u32 uid;
     u32 af;
     char task[TASK_COMM_LEN];
@@ -623,6 +759,7 @@ BPF_PERF_OUTPUT(ipv4_events);
 struct ipv6_event_t {
     u64 ts_us;
     u32 pid;
+    u32 ppid;
     u32 uid;
     u32 af;
     char task[TASK_COMM_LEN];
@@ -634,12 +771,33 @@ BPF_PERF_OUTPUT(ipv6_events);
 struct other_socket_event_t {
     u64 ts_us;
     u32 pid;
+    u32 ppid;
     u32 uid;
     u32 af;
     char task[TASK_COMM_LEN];
 } __attribute__((packed));
 BPF_PERF_OUTPUT(other_socket_events);
 
+// gethostlatency structs
+
+struct val_t {
+    u32 pid;
+    char comm[TASK_COMM_LEN];
+    char host[80];
+};
+
+struct data_dns_t {
+    u32 pid;
+    u32 ppid;
+    char comm[TASK_COMM_LEN];
+    char host[80];
+};
+
+BPF_HASH(start, u32, struct val_t);
+BPF_PERF_OUTPUT(dns_events);
+
+// execsnoop functions
+
 static int __submit_arg(struct pt_regs *ctx, void *ptr, struct data_t *data)
 {
     bpf_probe_read_user(data->argv, sizeof(data->argv), ptr);
@@ -720,8 +878,11 @@ int do_ret_sys_execve(struct pt_regs *ctx)
     return 0;
 }
 
+// securitySocketConnect functions
+
 int security_socket_connect_entry(struct pt_regs *ctx, struct socket *sock, struct sockaddr *address, int addrlen)
 {
+    struct task_struct *task;
     int ret = PT_REGS_RC(ctx);
 
     u64 pid_tgid = bpf_get_current_pid_tgid();
@@ -729,13 +890,16 @@ int security_socket_connect_entry(struct pt_regs *ctx, struct socket *sock, stru
 
     u32 uid = bpf_get_current_uid_gid();
 
+    task = (struct task_struct *)bpf_get_current_task();
+    u32 ppid = task->real_parent->tgid;
+
     struct sock *skp = sock->sk;
 
     // The AF options are listed in https://github.com/torvalds/linux/blob/master/include/linux/socket.h
 
     u32 address_family = address->sa_family;
     if (address_family == AF_INET) {
-        struct ipv4_event_t data4 = {.pid = pid, .uid = uid, .af = address_family};
+        struct ipv4_event_t data4 = {.pid = pid, .ppid = ppid, .uid = uid, .af = address_family};
         data4.ts_us = bpf_ktime_get_ns() / 1000;
 
         struct sockaddr_in *daddr = (struct sockaddr_in *)address;
@@ -753,7 +917,7 @@ int security_socket_connect_entry(struct pt_regs *ctx, struct socket *sock, stru
         }
     }
     else if (address_family == AF_INET6) {
-        struct ipv6_event_t data6 = {.pid = pid, .uid = uid, .af = address_family};
+        struct ipv6_event_t data6 = {.pid = pid, .ppid = ppid, .uid = uid, .af = address_family};
         data6.ts_us = bpf_ktime_get_ns() / 1000;
 
         struct sockaddr_in6 *daddr6 = (struct sockaddr_in6 *)address;
@@ -771,7 +935,7 @@ int security_socket_connect_entry(struct pt_regs *ctx, struct socket *sock, stru
         }
     }
     else if (address_family != AF_UNIX && address_family != AF_UNSPEC) { // other sockets, except UNIX and UNSPEC sockets
-        struct other_socket_event_t socket_event = {.pid = pid, .uid = uid, .af = address_family};
+        struct other_socket_event_t socket_event = {.pid = pid, .ppid = ppid, .uid = uid, .af = address_family};
         socket_event.ts_us = bpf_ktime_get_ns() / 1000;
         bpf_get_current_comm(&socket_event.task, sizeof(socket_event.task));
         other_socket_events.perf_submit(ctx, &socket_event, sizeof(socket_event));
@@ -779,6 +943,40 @@ int security_socket_connect_entry(struct pt_regs *ctx, struct socket *sock, stru
 
     return 0;
 }
+
+// gethostlatency functions
+
+int do_entry(struct pt_regs *ctx) {
+    if (!PT_REGS_PARM1(ctx))
+        return 0;
+    struct val_t val = {};
+    u32 pid = bpf_get_current_pid_tgid();
+    if (bpf_get_current_comm(&val.comm, sizeof(val.comm)) == 0) {
+        bpf_probe_read_user(&val.host, sizeof(val.host),
+                       (void *)PT_REGS_PARM1(ctx));
+        val.pid = pid;
+        start.update(&pid, &val);
+    }
+    return 0;
+}
+
+int do_return(struct pt_regs *ctx) {
+    struct val_t *valp;
+    struct data_dns_t data = {};
+    struct task_struct *task;
+    u32 pid = bpf_get_current_pid_tgid();
+    valp = start.lookup(&pid);
+    if (valp == 0)
+        return 0;       // missed start
+    bpf_probe_read_kernel(&data.comm, sizeof(data.comm), valp->comm);
+    bpf_probe_read_kernel(&data.host, sizeof(data.host), (void *)valp->host);
+    data.pid = valp->pid;
+    task = (struct task_struct *)bpf_get_current_task();
+    data.ppid = task->real_parent->tgid;
+    dns_events.perf_submit(ctx, &data, sizeof(data));
+    start.delete(&pid);
+    return 0;
+}
 """
 
 if __name__ == "__main__":
